@@ -1,14 +1,21 @@
 use crate::check::inappropriate_handshake_message;
-use crate::conn::{CommonState, ConnectionRandoms, State};
+#[cfg(feature = "quic")]
+use crate::check::inappropriate_message;
+#[cfg(feature = "quic")]
+use crate::common_state::Protocol;
+#[cfg(feature = "secret_extraction")]
+use crate::common_state::Side;
+use crate::common_state::{CommonState, State};
+use crate::conn::ConnectionRandoms;
 use crate::enums::ProtocolVersion;
-use crate::error::Error;
+use crate::enums::{AlertDescription, ContentType, HandshakeType};
+use crate::error::{Error, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::HandshakeHash;
 use crate::key::Certificate;
 #[cfg(feature = "logging")]
 use crate::log::{debug, trace, warn};
 use crate::msgs::codec::Codec;
-use crate::msgs::enums::{AlertDescription, KeyUpdateRequest};
-use crate::msgs::enums::{ContentType, HandshakeType};
+use crate::msgs::enums::KeyUpdateRequest;
 use crate::msgs::handshake::HandshakeMessagePayload;
 use crate::msgs::handshake::HandshakePayload;
 use crate::msgs::handshake::{NewSessionTicketExtension, NewSessionTicketPayloadTLS13};
@@ -16,14 +23,12 @@ use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
 use crate::rand;
 use crate::server::ServerConfig;
+#[cfg(feature = "secret_extraction")]
+use crate::suites::PartiallyExtractedSecrets;
 use crate::ticketer;
 use crate::tls13::key_schedule::{KeyScheduleTraffic, KeyScheduleTrafficWithClientFinishedPending};
 use crate::tls13::Tls13CipherSuite;
 use crate::verify;
-#[cfg(feature = "quic")]
-use crate::{check::inappropriate_message, conn::Protocol};
-#[cfg(feature = "secret_extraction")]
-use crate::{conn::Side, suites::PartiallyExtractedSecrets};
 
 use super::hs::{self, HandshakeHashOrBuffer, ServerContext};
 use super::server_conn::ServerConnectionData;
@@ -48,21 +53,19 @@ mod client_hello {
     use crate::msgs::handshake::CertificateRequestPayloadTLS13;
     use crate::msgs::handshake::CertificateStatus;
     use crate::msgs::handshake::ClientHelloPayload;
-    use crate::msgs::handshake::DigitallySignedStruct;
     use crate::msgs::handshake::HelloRetryExtension;
     use crate::msgs::handshake::HelloRetryRequest;
     use crate::msgs::handshake::KeyShareEntry;
     use crate::msgs::handshake::Random;
     use crate::msgs::handshake::ServerExtension;
     use crate::msgs::handshake::ServerHelloPayload;
-    use crate::msgs::handshake::SessionID;
-    #[cfg(feature = "quic")]
-    use crate::quic;
+    use crate::msgs::handshake::SessionId;
     use crate::server::common::ActiveCertifiedKey;
     use crate::sign;
     use crate::tls13::key_schedule::{
         KeyScheduleEarly, KeyScheduleHandshake, KeySchedulePreHandshake,
     };
+    use crate::verify::DigitallySignedStruct;
 
     use super::*;
 
@@ -79,7 +82,7 @@ mod client_hello {
         pub(in crate::server) suite: &'static Tls13CipherSuite,
         pub(in crate::server) randoms: ConnectionRandoms,
         pub(in crate::server) done_retry: bool,
-        pub(in crate::server) send_ticket: bool,
+        pub(in crate::server) send_tickets: usize,
         pub(in crate::server) extra_exts: Vec<ServerExtension>,
     }
 
@@ -117,7 +120,7 @@ mod client_hello {
                 .transcript
                 .get_hash_given(&binder_plaintext);
 
-            let key_schedule = KeyScheduleEarly::new(suite.hkdf_algorithm, psk);
+            let key_schedule = KeyScheduleEarly::new(suite, psk);
             let real_binder =
                 key_schedule.resumption_psk_binder_key_and_sign_verify_data(&handshake_hash);
 
@@ -132,12 +135,12 @@ mod client_hello {
                 self.config
                     .ticketer
                     .decrypt(ticket)
-                    .and_then(|plain| persist::ServerSessionValue::read_bytes(&plain))
+                    .and_then(|plain| persist::ServerSessionValue::read_bytes(&plain).ok())
             } else {
                 self.config
                     .session_storage
                     .take(ticket)
-                    .and_then(|plain| persist::ServerSessionValue::read_bytes(&plain))
+                    .and_then(|plain| persist::ServerSessionValue::read_bytes(&plain).ok())
             }
         }
 
@@ -150,35 +153,50 @@ mod client_hello {
             mut sigschemes_ext: Vec<SignatureScheme>,
         ) -> hs::NextStateOrError {
             if client_hello.compression_methods.len() != 1 {
-                return Err(cx
-                    .common
-                    .illegal_param("client offered wrong compressions"));
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::IllegalParameter,
+                    PeerMisbehaved::OfferedIncorrectCompressions,
+                ));
             }
 
             let groups_ext = client_hello
                 .get_namedgroups_extension()
-                .ok_or_else(|| hs::incompatible(cx.common, "client didn't describe groups"))?;
+                .ok_or_else(|| {
+                    cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::NamedGroupsExtensionRequired,
+                    )
+                })?;
 
             let tls13_schemes = sign::supported_sign_tls13();
             sigschemes_ext.retain(|scheme| tls13_schemes.contains(scheme));
 
             let shares_ext = client_hello
                 .get_keyshare_extension()
-                .ok_or_else(|| hs::incompatible(cx.common, "client didn't send keyshares"))?;
+                .ok_or_else(|| {
+                    cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::KeyShareExtensionRequired,
+                    )
+                })?;
 
             if client_hello.has_keyshare_extension_with_duplicates() {
-                return Err(cx
-                    .common
-                    .illegal_param("client sent duplicate keyshares"));
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::IllegalParameter,
+                    PeerMisbehaved::OfferedDuplicateKeyShares,
+                ));
             }
 
             let early_data_requested = client_hello.early_data_extension_offered();
 
             // EarlyData extension is illegal in second ClientHello
             if self.done_retry && early_data_requested {
-                return Err(cx
-                    .common
-                    .illegal_param("client sent EarlyData in second ClientHello"));
+                return Err({
+                    cx.common.send_fatal_alert(
+                        AlertDescription::IllegalParameter,
+                        PeerMisbehaved::EarlyDataAttemptedInSecondClientHello,
+                    )
+                });
             }
 
             // choose a share that we support
@@ -208,9 +226,10 @@ mod client_hello {
 
                     if let Some(group) = retry_group_maybe {
                         if self.done_retry {
-                            return Err(cx
-                                .common
-                                .illegal_param("did not follow retry request"));
+                            return Err(cx.common.send_fatal_alert(
+                                AlertDescription::IllegalParameter,
+                                PeerMisbehaved::RefusedToFollowHelloRetryRequest,
+                            ));
                         }
 
                         emit_hello_retry_request(
@@ -227,11 +246,11 @@ mod client_hello {
                             config: self.config,
                             transcript: HandshakeHashOrBuffer::Hash(self.transcript),
                             #[cfg(feature = "tls12")]
-                            session_id: SessionID::empty(),
+                            session_id: SessionId::empty(),
                             #[cfg(feature = "tls12")]
                             using_ems: false,
                             done_retry: true,
-                            send_ticket: self.send_ticket,
+                            send_tickets: self.send_tickets,
                             extra_exts: self.extra_exts,
                         });
 
@@ -245,9 +264,9 @@ mod client_hello {
                         };
                     }
 
-                    return Err(hs::incompatible(
-                        cx.common,
-                        "no kx group overlap with client",
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::HandshakeFailure,
+                        PeerIncompatible::NoKxGroupsInCommon,
                     ));
                 }
             };
@@ -258,19 +277,35 @@ mod client_hello {
 
             if let Some(psk_offer) = client_hello.get_psk() {
                 if !client_hello.check_psk_ext_is_last() {
-                    return Err(cx
-                        .common
-                        .illegal_param("psk extension in wrong position"));
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::IllegalParameter,
+                        PeerMisbehaved::PskExtensionMustBeLast,
+                    ));
+                }
+
+                // "A client MUST provide a "psk_key_exchange_modes" extension if it
+                //  offers a "pre_shared_key" extension. If clients offer
+                //  "pre_shared_key" without a "psk_key_exchange_modes" extension,
+                //  servers MUST abort the handshake." - RFC8446 4.2.9
+                if client_hello.get_psk_modes().is_none() {
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::MissingExtension,
+                        PeerMisbehaved::MissingPskModesExtension,
+                    ));
                 }
 
                 if psk_offer.binders.is_empty() {
-                    return Err(hs::decode_error(cx.common, "psk extension missing binder"));
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::DecodeError,
+                        PeerMisbehaved::MissingBinderInPskExtension,
+                    ));
                 }
 
                 if psk_offer.binders.len() != psk_offer.identities.len() {
-                    return Err(cx
-                        .common
-                        .illegal_param("psk extension mismatched ids/binders"));
+                    return Err(cx.common.send_fatal_alert(
+                        AlertDescription::IllegalParameter,
+                        PeerMisbehaved::PskExtensionWithMismatchedIdsAndBinders,
+                    ));
                 }
 
                 for (i, psk_id) in psk_offer.identities.iter().enumerate() {
@@ -290,12 +325,11 @@ mod client_hello {
                         self.suite,
                         chm,
                         &resume.master_secret.0,
-                        &psk_offer.binders[i].0,
+                        psk_offer.binders[i].as_ref(),
                     ) {
-                        cx.common
-                            .send_fatal_alert(AlertDescription::DecryptError);
-                        return Err(Error::PeerMisbehavedError(
-                            "client sent wrong binder".to_string(),
+                        return Err(cx.common.send_fatal_alert(
+                            AlertDescription::DecryptError,
+                            PeerMisbehaved::IncorrectBinder,
                         ));
                     }
 
@@ -307,11 +341,11 @@ mod client_hello {
 
             if !client_hello.psk_mode_offered(PSKKeyExchangeMode::PSK_DHE_KE) {
                 debug!("Client unwilling to resume, DHE_KE not offered");
-                self.send_ticket = false;
+                self.send_tickets = 0;
                 chosen_psk_index = None;
                 resumedata = None;
             } else {
-                self.send_ticket = true;
+                self.send_tickets = self.config.send_tls13_tickets;
             }
 
             if let Some(ref resume) = resumedata {
@@ -377,23 +411,15 @@ mod client_hello {
             // are encrypted with the handshake keys.
             match doing_early_data {
                 EarlyDataDecision::Disabled => {
-                    cx.common
-                        .record_layer
-                        .set_message_decrypter(
-                            self.suite
-                                .derive_decrypter(key_schedule.client_key()),
-                        );
+                    key_schedule.set_handshake_decrypter(None, cx.common);
                     cx.data.early_data.reject();
                 }
                 EarlyDataDecision::RequestedButRejected => {
                     debug!("Client requested early_data, but not accepted: switching to handshake keys with trial decryption");
-                    cx.common
-                        .record_layer
-                        .set_message_decrypter_with_trial_decryption(
-                            self.suite
-                                .derive_decrypter(key_schedule.client_key()),
-                            max_early_data_size(self.config.max_early_data_size),
-                        );
+                    key_schedule.set_handshake_decrypter(
+                        Some(max_early_data_size(self.config.max_early_data_size)),
+                        cx.common,
+                    );
                     cx.data.early_data.reject();
                 }
                 EarlyDataDecision::Accepted => {
@@ -406,7 +432,6 @@ mod client_hello {
             cx.common.check_aligned_handshake()?;
             let key_schedule_traffic = emit_finished_tls13(
                 &mut self.transcript,
-                self.suite,
                 &self.randoms,
                 cx,
                 key_schedule,
@@ -426,7 +451,7 @@ mod client_hello {
                     transcript: self.transcript,
                     suite: self.suite,
                     key_schedule: key_schedule_traffic,
-                    send_ticket: self.send_ticket,
+                    send_tickets: self.send_tickets,
                 }))
             } else if doing_early_data == EarlyDataDecision::Accepted && !cx.common.is_quic() {
                 // Not used for QUIC: RFC 9001 §8.3: Clients MUST NOT send the EndOfEarlyData
@@ -437,7 +462,7 @@ mod client_hello {
                     transcript: self.transcript,
                     suite: self.suite,
                     key_schedule: key_schedule_traffic,
-                    send_ticket: self.send_ticket,
+                    send_tickets: self.send_tickets,
                 }))
             } else {
                 Ok(Box::new(ExpectFinished {
@@ -445,7 +470,7 @@ mod client_hello {
                     transcript: self.transcript,
                     suite: self.suite,
                     key_schedule: key_schedule_traffic,
-                    send_ticket: self.send_ticket,
+                    send_tickets: self.send_tickets,
                 }))
             }
         }
@@ -456,7 +481,7 @@ mod client_hello {
         randoms: &ConnectionRandoms,
         suite: &'static Tls13CipherSuite,
         cx: &mut ServerContext<'_>,
-        session_id: &SessionID,
+        session_id: &SessionId,
         share: &KeyShareEntry,
         chosen_psk_idx: Option<usize>,
         resuming_psk: Option<&[u8]>,
@@ -501,20 +526,18 @@ mod client_hello {
         cx.common.send_msg(sh, false);
 
         // Start key schedule
-        let (key_schedule_pre_handshake, early_data_client_key) = if let Some(psk) = resuming_psk {
-            let early_key_schedule = KeyScheduleEarly::new(suite.hkdf_algorithm, psk);
-            let client_early_traffic_secret = early_key_schedule.client_early_traffic_secret(
+        let key_schedule_pre_handshake = if let Some(psk) = resuming_psk {
+            let early_key_schedule = KeyScheduleEarly::new(suite, psk);
+            early_key_schedule.client_early_traffic_secret(
                 &client_hello_hash,
                 &*config.key_log,
                 &randoms.client,
+                cx.common,
             );
 
-            (
-                KeySchedulePreHandshake::from(early_key_schedule),
-                Some(client_early_traffic_secret),
-            )
+            KeySchedulePreHandshake::from(early_key_schedule)
         } else {
-            (KeySchedulePreHandshake::new(suite.hkdf_algorithm), None)
+            KeySchedulePreHandshake::new(suite)
         };
 
         // Do key exchange
@@ -523,34 +546,13 @@ mod client_hello {
         })?;
 
         let handshake_hash = transcript.get_current_hash();
-        let (key_schedule, _client_key, server_key) = key_schedule.derive_handshake_secrets(
+        let key_schedule = key_schedule.derive_server_handshake_secrets(
             handshake_hash,
             &*config.key_log,
             &randoms.client,
+            cx.common,
         );
 
-        // Set up to encrypt with handshake secrets, but decrypt with early_data keys.
-        // If not doing early_data after all, this is corrected later to the handshake
-        // keys (now stored in key_schedule).
-        cx.common
-            .record_layer
-            .set_message_encrypter(suite.derive_encrypter(&server_key));
-
-        if let Some(key) = &early_data_client_key {
-            cx.common
-                .record_layer
-                .set_message_decrypter(suite.derive_decrypter(key));
-        }
-
-        #[cfg(feature = "quic")]
-        if cx.common.is_quic() {
-            // If 0-RTT should be rejected, this will be clobbered by ExtensionProcessing
-            // before the application can see.
-            cx.common.quic.early_secret = early_data_client_key;
-            cx.common.quic.hs_secrets =
-                Some(quic::Secrets::new(_client_key, server_key, suite, false));
-        }
-
         Ok(key_schedule)
     }
 
@@ -573,7 +575,7 @@ mod client_hello {
     ) {
         let mut req = HelloRetryRequest {
             legacy_version: ProtocolVersion::TLSv1_2,
-            session_id: SessionID::empty(),
+            session_id: SessionId::empty(),
             cipher_suite: suite.common.suite,
             extensions: Vec::new(),
         };
@@ -599,6 +601,7 @@ mod client_hello {
         common.send_msg(m, false);
     }
 
+    #[allow(clippy::needless_pass_by_ref_mut)] // cx only mutated if cfg(feature = "quic")
     fn decide_if_early_data_allowed(
         cx: &mut ServerContext<'_>,
         client_hello: &ClientHelloPayload,
@@ -723,12 +726,7 @@ mod client_hello {
         let names = config
             .verifier
             .client_auth_root_subjects()
-            .ok_or_else(|| {
-                debug!("could not determine root subjects based on SNI");
-                cx.common
-                    .send_fatal_alert(AlertDescription::AccessDenied);
-                Error::General("client rejected by client_auth_root_subjects".into())
-            })?;
+            .to_vec();
 
         if !names.is_empty() {
             cr.extensions
@@ -808,7 +806,12 @@ mod client_hello {
 
         let signer = signing_key
             .choose_scheme(schemes)
-            .ok_or_else(|| hs::incompatible(common, "no overlapping sigschemes"))?;
+            .ok_or_else(|| {
+                common.send_fatal_alert(
+                    AlertDescription::HandshakeFailure,
+                    PeerIncompatible::NoSignatureSchemesInCommon,
+                )
+            })?;
 
         let scheme = signer.scheme();
         let sig = signer.sign(&message)?;
@@ -831,7 +834,6 @@ mod client_hello {
 
     fn emit_finished_tls13(
         transcript: &mut HandshakeHash,
-        suite: &'static Tls13CipherSuite,
         randoms: &ConnectionRandoms,
         cx: &mut ServerContext<'_>,
         key_schedule: KeyScheduleHandshake,
@@ -856,23 +858,12 @@ mod client_hello {
 
         // Now move to application data keys.  Read key change is deferred until
         // the Finish message is received & validated.
-        let (key_schedule_traffic, _client_key, server_key) = key_schedule
-            .into_traffic_with_client_finished_pending(
-                hash_at_server_fin,
-                &*config.key_log,
-                &randoms.client,
-            );
-        cx.common
-            .record_layer
-            .set_message_encrypter(suite.derive_encrypter(&server_key));
-
-        #[cfg(feature = "quic")]
-        {
-            cx.common.quic.traffic_secrets =
-                Some(quic::Secrets::new(_client_key, server_key, suite, false));
-        }
-
-        key_schedule_traffic
+        key_schedule.into_traffic_with_client_finished_pending(
+            hash_at_server_fin,
+            &*config.key_log,
+            &randoms.client,
+            cx.common,
+        )
     }
 }
 
@@ -903,7 +894,7 @@ struct ExpectCertificate {
     transcript: HandshakeHash,
     suite: &'static Tls13CipherSuite,
     key_schedule: KeyScheduleTrafficWithClientFinishedPending,
-    send_ticket: bool,
+    send_tickets: usize,
 }
 
 impl State<ServerConnectionData> for ExpectCertificate {
@@ -918,9 +909,7 @@ impl State<ServerConnectionData> for ExpectCertificate {
         // We don't send any CertificateRequest extensions, so any extensions
         // here are illegal.
         if certp.any_entry_has_extension() {
-            return Err(Error::PeerMisbehavedError(
-                "client sent unsolicited cert extension".to_string(),
-            ));
+            return Err(PeerMisbehaved::UnsolicitedCertExtension.into());
         }
 
         let client_cert = certp.convert();
@@ -928,13 +917,7 @@ impl State<ServerConnectionData> for ExpectCertificate {
         let mandatory = self
             .config
             .verifier
-            .client_auth_mandatory()
-            .ok_or_else(|| {
-                debug!("could not determine if client auth is mandatory based on SNI");
-                cx.common
-                    .send_fatal_alert(AlertDescription::AccessDenied);
-                Error::General("client rejected by client_auth_mandatory".into())
-            })?;
+            .client_auth_mandatory();
 
         let (end_entity, intermediates) = match client_cert.split_first() {
             None => {
@@ -946,13 +929,14 @@ impl State<ServerConnectionData> for ExpectCertificate {
                         suite: self.suite,
                         key_schedule: self.key_schedule,
                         transcript: self.transcript,
-                        send_ticket: self.send_ticket,
+                        send_tickets: self.send_tickets,
                     }));
                 }
 
-                cx.common
-                    .send_fatal_alert(AlertDescription::CertificateRequired);
-                return Err(Error::NoCertificatesPresented);
+                return Err(cx.common.send_fatal_alert(
+                    AlertDescription::CertificateRequired,
+                    Error::NoCertificatesPresented,
+                ));
             }
             Some(chain) => chain,
         };
@@ -962,8 +946,8 @@ impl State<ServerConnectionData> for ExpectCertificate {
             .verifier
             .verify_client_cert(end_entity, intermediates, now)
             .map_err(|err| {
-                hs::incompatible(cx.common, "certificate invalid");
-                err
+                cx.common
+                    .send_cert_verify_error_alert(err)
             })?;
 
         Ok(Box::new(ExpectCertificateVerify {
@@ -972,7 +956,7 @@ impl State<ServerConnectionData> for ExpectCertificate {
             transcript: self.transcript,
             key_schedule: self.key_schedule,
             client_cert,
-            send_ticket: self.send_ticket,
+            send_tickets: self.send_tickets,
         }))
     }
 }
@@ -983,7 +967,7 @@ struct ExpectCertificateVerify {
     suite: &'static Tls13CipherSuite,
     key_schedule: KeyScheduleTrafficWithClientFinishedPending,
     client_cert: Vec<Certificate>,
-    send_ticket: bool,
+    send_tickets: usize,
 }
 
 impl State<ServerConnectionData> for ExpectCertificateVerify {
@@ -1005,9 +989,9 @@ impl State<ServerConnectionData> for ExpectCertificateVerify {
         };
 
         if let Err(e) = rc {
-            cx.common
-                .send_fatal_alert(AlertDescription::AccessDenied);
-            return Err(e);
+            return Err(cx
+                .common
+                .send_cert_verify_error_alert(e));
         }
 
         trace!("client CertificateVerify OK");
@@ -1019,7 +1003,7 @@ impl State<ServerConnectionData> for ExpectCertificateVerify {
             suite: self.suite,
             key_schedule: self.key_schedule,
             transcript: self.transcript,
-            send_ticket: self.send_ticket,
+            send_tickets: self.send_tickets,
         }))
     }
 }
@@ -1032,7 +1016,7 @@ struct ExpectEarlyData {
     transcript: HandshakeHash,
     suite: &'static Tls13CipherSuite,
     key_schedule: KeyScheduleTrafficWithClientFinishedPending,
-    send_ticket: bool,
+    send_tickets: usize,
 }
 
 impl State<ServerConnectionData> for ExpectEarlyData {
@@ -1045,13 +1029,10 @@ impl State<ServerConnectionData> for ExpectEarlyData {
                     .take_received_plaintext(payload)
                 {
                     true => Ok(self),
-                    false => {
-                        cx.common
-                            .send_fatal_alert(AlertDescription::UnexpectedMessage);
-                        Err(Error::PeerMisbehavedError(
-                            "too much early_data received".into(),
-                        ))
-                    }
+                    false => Err(cx.common.send_fatal_alert(
+                        AlertDescription::UnexpectedMessage,
+                        PeerMisbehaved::TooMuchEarlyDataReceived,
+                    )),
                 }
             }
             MessagePayload::Handshake {
@@ -1062,20 +1043,15 @@ impl State<ServerConnectionData> for ExpectEarlyData {
                     },
                 ..
             } => {
-                cx.common
-                    .record_layer
-                    .set_message_decrypter(
-                        self.suite
-                            .derive_decrypter(self.key_schedule.client_key()),
-                    );
-
+                self.key_schedule
+                    .update_decrypter(cx.common);
                 self.transcript.add_message(&m);
                 Ok(Box::new(ExpectFinished {
                     config: self.config,
                     suite: self.suite,
                     key_schedule: self.key_schedule,
                     transcript: self.transcript,
-                    send_ticket: self.send_ticket,
+                    send_tickets: self.send_tickets,
                 }))
             }
             payload => Err(inappropriate_handshake_message(
@@ -1089,7 +1065,7 @@ impl State<ServerConnectionData> for ExpectEarlyData {
 
 // --- Process client's Finished ---
 fn get_server_session_value(
-    transcript: &mut HandshakeHash,
+    transcript: &HandshakeHash,
     suite: &'static Tls13CipherSuite,
     key_schedule: &KeyScheduleTraffic,
     cx: &ServerContext<'_>,
@@ -1121,12 +1097,12 @@ struct ExpectFinished {
     transcript: HandshakeHash,
     suite: &'static Tls13CipherSuite,
     key_schedule: KeyScheduleTrafficWithClientFinishedPending,
-    send_ticket: bool,
+    send_tickets: usize,
 }
 
 impl ExpectFinished {
     fn emit_ticket(
-        transcript: &mut HandshakeHash,
+        transcript: &HandshakeHash,
         suite: &'static Tls13CipherSuite,
         cx: &mut ServerContext<'_>,
         key_schedule: &KeyScheduleTraffic,
@@ -1184,7 +1160,6 @@ impl ExpectFinished {
         };
 
         trace!("sending new ticket {:?} (stateless: {})", m, stateless);
-        transcript.add_message(&m);
         cx.common.send_msg(m, true);
         Ok(())
     }
@@ -1196,16 +1171,15 @@ impl State<ServerConnectionData> for ExpectFinished {
             require_handshake_msg!(m, HandshakeType::Finished, HandshakePayload::Finished)?;
 
         let handshake_hash = self.transcript.get_current_hash();
-        let (key_schedule_traffic, expect_verify_data, client_key) = self
+        let (key_schedule_traffic, expect_verify_data) = self
             .key_schedule
-            .sign_client_finish(&handshake_hash);
+            .sign_client_finish(&handshake_hash, cx.common);
 
         let fin = constant_time::verify_slices_are_equal(expect_verify_data.as_ref(), &finished.0)
             .map_err(|_| {
-                cx.common
-                    .send_fatal_alert(AlertDescription::DecryptError);
                 warn!("Finished wrong");
-                Error::DecryptError
+                cx.common
+                    .send_fatal_alert(AlertDescription::DecryptError, Error::DecryptError)
             })
             .map(|_| verify::FinishedMessageVerified::assertion())?;
 
@@ -1215,14 +1189,9 @@ impl State<ServerConnectionData> for ExpectFinished {
 
         cx.common.check_aligned_handshake()?;
 
-        // Install keying to read future messages.
-        cx.common
-            .record_layer
-            .set_message_decrypter(self.suite.derive_decrypter(&client_key));
-
-        if self.send_ticket {
+        for _ in 0..self.send_tickets {
             Self::emit_ticket(
-                &mut self.transcript,
+                &self.transcript,
                 self.suite,
                 cx,
                 &key_schedule_traffic,
@@ -1244,9 +1213,7 @@ impl State<ServerConnectionData> for ExpectFinished {
         }
 
         Ok(Box::new(ExpectTraffic {
-            suite: self.suite,
             key_schedule: key_schedule_traffic,
-            want_write_key_update: false,
             _fin_verified: fin,
         }))
     }
@@ -1254,9 +1221,7 @@ impl State<ServerConnectionData> for ExpectFinished {
 
 // --- Process traffic ---
 struct ExpectTraffic {
-    suite: &'static Tls13CipherSuite,
     key_schedule: KeyScheduleTraffic,
-    want_write_key_update: bool,
     _fin_verified: verify::FinishedMessageVerified,
 }
 
@@ -1264,42 +1229,28 @@ impl ExpectTraffic {
     fn handle_key_update(
         &mut self,
         common: &mut CommonState,
-        kur: &KeyUpdateRequest,
+        key_update_request: &KeyUpdateRequest,
     ) -> Result<(), Error> {
         #[cfg(feature = "quic")]
         {
             if let Protocol::Quic = common.protocol {
-                common.send_fatal_alert(AlertDescription::UnexpectedMessage);
-                let msg = "KeyUpdate received in QUIC connection".to_string();
-                warn!("{}", msg);
-                return Err(Error::PeerMisbehavedError(msg));
+                return Err(common.send_fatal_alert(
+                    AlertDescription::UnexpectedMessage,
+                    PeerMisbehaved::KeyUpdateReceivedInQuicConnection,
+                ));
             }
         }
 
         common.check_aligned_handshake()?;
 
-        match kur {
-            KeyUpdateRequest::UpdateNotRequested => {}
-            KeyUpdateRequest::UpdateRequested => {
-                self.want_write_key_update = true;
-            }
-            _ => {
-                common.send_fatal_alert(AlertDescription::IllegalParameter);
-                return Err(Error::CorruptMessagePayload(ContentType::Handshake));
-            }
+        if common.should_update_key(key_update_request)? {
+            self.key_schedule
+                .update_encrypter_and_notify(common);
         }
 
         // Update our read-side keys.
-        let new_read_key = self
-            .key_schedule
-            .next_client_application_traffic_secret();
-        common
-            .record_layer
-            .set_message_decrypter(
-                self.suite
-                    .derive_decrypter(&new_read_key),
-            );
-
+        self.key_schedule
+            .update_decrypter(common);
         Ok(())
     }
 }
@@ -1340,24 +1291,10 @@ impl State<ServerConnectionData> for ExpectTraffic {
             .export_keying_material(output, label, context)
     }
 
-    fn perhaps_write_key_update(&mut self, common: &mut CommonState) {
-        if self.want_write_key_update {
-            self.want_write_key_update = false;
-            common.send_msg_encrypt(Message::build_key_update_notify().into());
-
-            let write_key = self
-                .key_schedule
-                .next_server_application_traffic_secret();
-            common
-                .record_layer
-                .set_message_encrypter(self.suite.derive_encrypter(&write_key));
-        }
-    }
-
     #[cfg(feature = "secret_extraction")]
     fn extract_secrets(&self) -> Result<PartiallyExtractedSecrets, Error> {
         self.key_schedule
-            .extract_secrets(self.suite.common.aead_algorithm, Side::Server)
+            .extract_secrets(Side::Server)
     }
 }
 
